#include <stdio.h>
#include <stdlib.h>
#include <portaudio.h>

#include "ugens/ugens.h"

typedef struct
{
    graph_t *graph;
} paData;

static int paCallback( const void *inputBuffer, void *outputBuffer,
                           unsigned long framesPerBuffer,
                           const PaStreamCallbackTimeInfo* timeInfo,
                           PaStreamCallbackFlags statusFlags,
                           void *userData )
{
    /* Cast data passed through stream to our structure. */
    paData *data = (paData*)userData;
    float *out = (float*)outputBuffer;
//    float *in = (float*)inputBuffer;
    unsigned int i;
    (void) inputBuffer; /* Prevent unused variable warning. */
    for( i=0; i<framesPerBuffer; i++ ) {
//        graph_node_t *node = data->graphHead;
        double *graph_out = graph_get_next(data->graph);
//        while (node) {
//            node->tick(node->ugen_instance);
//            node = node->next;
//        }
        *out++ = graph_out[0];
        *out++ = graph_out[1];
    }
    return paNoError;
}



int main(void)
{
    PaError err;
    int sr = 44100;
    paData data;

    double freq = 440;

    graph_t *graph = data.graph = graph_new(0, 2);

    graph_node_t *env = graph_create_node(graph, "linen", sr, 1.0, 3.0, 1.0, 2.0);
    graph_node_t *filterenv = graph_create_node(graph, "linen", sr, 6000.0, 1.0, 1.0, 4.0);
    graph_node_t *osc = graph_create_node(graph, "sinosc", sr, freq, 0.5);
    graph_node_t *shaper = graph_create_node(graph, "wshaper");
    graph_node_t *fenv_offset = graph_create_node(graph, "add");
    graph_node_t *filter = graph_create_node(graph, "filter", sr);
    
    graph_connect(graph, env, 0, shaper, WSHAPER_MORPH_RATIO);
    graph_connect(graph, osc, 0, shaper, WSHAPER_IN);
    graph_connect(graph, shaper, 0, filter, FILTER_IN);
    graph_connect(graph, filterenv, 0, fenv_offset, ADD_IN1);
    graph_connect(graph, fenv_offset, 0, filter, FILTER_CF);
    double *value = graph_get_ugen_input(graph, fenv_offset, 1);
    *value = 200;

    graph_set_out(graph, filter, 0, 0);

    // Set default values
    wshaper_set_param(shaper->ugen_instance, WSHAPER_DEPTH, 0.5);
    filter_set_param(filter->ugen_instance, FILTER_CF, 500);
    filter_set_param(filter->ugen_instance, FILTER_RES, 0.5);

    err = Pa_Initialize();
    if( err != paNoError )
        printf( "PortAudio error: %s\n", Pa_GetErrorText( err ) );

    PaStream *stream;

//    PaStreamParameters inputParameters;
//    PaStreamParameters outputParameters;

//    inputParameters.channelCount = 1;
//    inputParameters.device = 0;
//    inputParameters.hostApiSpecificStreamInfo = NULL;
//    inputParameters.sampleFormat = paFloat32;
//    inputParameters.suggestedLatency = 0.05;


//    outputParameters.channelCount = 1;
//    outputParameters.device = 0;
//    outputParameters.hostApiSpecificStreamInfo = NULL;
//    outputParameters.sampleFormat = paFloat32;
//    outputParameters.suggestedLatency = 0.05;


//    Pa_OpenStream(&stream,
//                  &inputParameters,
//                  &outputParameters,
//                  sr,
//                  128, /* frames per buffer */
//                  paNoFlag,
//                  paCallback, /* this is your callback function */
//                  &data);


    err = Pa_OpenDefaultStream( &stream,
                                2, /* no input channels */
                                2, /* stereo output */
                                paFloat32, /* 32 bit floating point output */
                                sr,
                                128, /* frames per buffer */
                                paCallback, /* this is your callback function */
                                &data ); /*This is a pointer that will be passed to
   your callback*/
    if( err != paNoError )
        printf( "PortAudio error: %s\n", Pa_GetErrorText( err ) );

    err = Pa_StartStream( stream );
    if( err != paNoError )
        printf( "PortAudio error: %s\n", Pa_GetErrorText( err ) );

    Pa_Sleep(100*1000);

    err = Pa_StopStream( stream );
    if( err != paNoError )
        printf( "PortAudio error: %s\n", Pa_GetErrorText( err ) );

    err= Pa_CloseStream(stream);
    if( err != paNoError )
        printf( "PortAudio error: %s\n", Pa_GetErrorText( err ) );

    err = Pa_Terminate();
    if( err != paNoError )
        printf( "PortAudio error: %s\n", Pa_GetErrorText( err ) );

    printf("Hello World!\n");
    return 0;
}

